import os
import yaml
import argh
import torch
import pkbar
from dagshub import DAGsHubLogger
import torchvision
from model import get_instance_segmentation_model
from dataset import CocoDetection, get_transforms
from ignite.engine import Events, Engine, create_supervised_evaluator, convert_tensor
from ignite.metrics import Loss
from ignite.handlers import EarlyStopping
from torch.nn import functional as F


# Monkey path _resize_image_and_masks to prevent warnings
def _resize_image_and_masks(image, self_min_size, self_max_size, target):
    # type: (Tensor, float, float, Optional[Dict[str, Tensor]]) -> Tuple[Tensor, Optional[Dict[str, Tensor]]]
    im_shape = torch.tensor(image.shape[-2:])
    min_size = float(torch.min(im_shape))
    max_size = float(torch.max(im_shape))
    scale_factor = self_min_size / min_size
    if max_size * scale_factor > self_max_size:
        scale_factor = self_max_size / max_size
    image = torch.nn.functional.interpolate(
        image[None], scale_factor=scale_factor, mode='bilinear', recompute_scale_factor=True,
        align_corners=False)[0]

    if target is None:
        return image, target

    if "masks" in target:
        mask = target["masks"]
        mask = F.interpolate(mask[:, None].float(), scale_factor=scale_factor, recompute_scale_factor=True)[:, 0].byte()
        target["masks"] = mask
    return image, target


torchvision.models.detection.transform._resize_image_and_masks = _resize_image_and_masks


def collate_fn(batch):
    return tuple(zip(*batch))


def train(
        dataset_path,
        model_path,
        batch_size=2,
        lr=0.001,
        lr_step_size=3,
        lr_gamma=0.1,
        op_momentum=0.9,
        op_weight_decay=0.0005,
        epochs=10,
        log_interval=1,
):

    train_path = os.path.join(dataset_path, "train")
    val_path = os.path.join(dataset_path, "test")

    # use our dataset and defined transformations
    dataset = CocoDetection(os.path.join(train_path, "images"), os.path.join(train_path, "annotations/annotations.json"), transforms=get_transforms(train=True, preview=False))
    dataset_val = CocoDetection(os.path.join(val_path, "images"), os.path.join(val_path, "annotations/annotations.json"), transforms=get_transforms(train=False))
    # split the dataset in train and test set
    torch.manual_seed(42)

    # define training and validation data loaders
    data_loader = torch.utils.data.DataLoader(
        dataset, batch_size=batch_size, shuffle=True, num_workers=4, collate_fn=collate_fn)
    data_loader_val = torch.utils.data.DataLoader(
        dataset_val, batch_size=1, shuffle=False, num_workers=4, collate_fn=collate_fn)

    device = torch.device('cuda') if torch.cuda.is_available() else torch.device('cpu')

    # our dataset has two classes only - background and person
    num_classes = 2

    # get the model using our helper function
    model = get_instance_segmentation_model(num_classes)
    # move model to the right device
    model.to(device)

    # construct an optimizer
    params = [p for p in model.parameters() if p.requires_grad]
    optimizer = torch.optim.SGD(params, lr=lr,
                                momentum=op_momentum,
                                weight_decay=op_weight_decay)

    train_loader, val_loader = data_loader, data_loader_val
    trainer = create_maskrcnn_trainer(model, optimizer, device)

    val_metrics = {
        "loss": Loss(loss_fn)
    }
    evaluator = create_maskrcnn_evaluator(model, metrics=val_metrics, device=device)
    csv_logger = DAGsHubLogger(should_log_hparams=False)

    @trainer.on(Events.EPOCH_STARTED)
    def init_pbar(trainer):
        trainer.kbar = pkbar.Kbar(
            target=len(data_loader),
            epoch=trainer.state.epoch-1,
            num_epochs=trainer.state.max_epochs,
            width=8,
            always_stateful=False)

    @trainer.on(Events.ITERATION_COMPLETED(every=log_interval))
    def log_iteration(trainer):
        pass

    @trainer.on(Events.ITERATION_COMPLETED(every=1))
    def advance_iteration_pbar(trainer):
        iteration = trainer.state.iteration - (trainer.state.epoch-1) * trainer.kbar.target
        trainer.kbar.update(iteration-1, values=list(trainer.state.output.items()))

    @trainer.on(Events.EPOCH_COMPLETED)
    def log_epoch(trainer):
        evaluator.run(val_loader)
        metrics = evaluator.state.metrics
        csv_logger.log_metrics(metrics, step_num=trainer.state.epoch)

        output = {key: float(val) for key, val in trainer.state.output.items()}
        csv_logger.log_metrics(output, step_num=trainer.state.epoch)

    @trainer.on(Events.EPOCH_COMPLETED)
    def advance_epoch_pbar(trainer):
        metrics = evaluator.state.metrics
        trainer.kbar.add(1, values=[("val_loss", metrics["loss"])])

    def score_function(engine):
        val_loss = engine.state.metrics['loss']
        return 1-val_loss

    handler = EarlyStopping(patience=3, score_function=score_function, trainer=trainer)
    evaluator.add_event_handler(Events.COMPLETED, handler)

    trainer.run(train_loader, max_epochs=epochs)
    os.makedirs(model_path, exist_ok=True)
    torch.save(model, "models/model.pth")
    csv_logger.save()
    csv_logger.close()


def create_maskrcnn_trainer(model, optimizer, device):
    def _update(engine, batch):
        x, y = batch
        model.train()
        loss_dict = model(convert_tensor(x, device=device, non_blocking=False), convert_tensor(y, device=device, non_blocking=False))
        losses = sum(loss for loss in loss_dict.values())
        optimizer.zero_grad()
        losses.backward()
        optimizer.step()

        return loss_dict

    trainer = Engine(_update)

    return trainer


def create_maskrcnn_evaluator(model, metrics, device):
    def _inference(engine, batch):
        with torch.no_grad():
            x, y = batch
            loss_dict = model(convert_tensor(x, device=device, non_blocking=False), convert_tensor(y, device=device, non_blocking=False))
            return (loss_dict, y)

    evaluator = Engine(_inference)

    for name, metric in metrics.items():
        metric.attach(evaluator, name)

    return evaluator








    # and a learning rate scheduler which decreases the learning rate by
    # lr_scheduler = torch.optim.lr_scheduler.StepLR(optimizer,
    #                                                step_size=lr_step_size,
    #                                                gamma=lr_gamma)
    #
    # for epoch in range(epochs):
    #     kbar = pkbar.Kbar(target=len(data_loader), epoch=epoch, num_epochs=epochs, width=8, always_stateful=False)
    #     # train for one epoch, printing every 10 iterations
    #     train_one_epoch(model, optimizer, data_loader, kbar)
    #     # update the learning rate
    #     lr_scheduler.step()


def loss_fn(pred, y=None):
    return sum(loss for loss in pred.values())


def train_one_epoch(model, optimizer, data_loader, kbar):
    model.train()
    for i, (images, targets) in enumerate(data_loader):

        loss_dict = model(images, targets)
        losses = sum(loss for loss in loss_dict.values())

        optimizer.zero_grad()
        losses.backward()
        optimizer.step()

        # if lr_scheduler is not None:
        #     lr_scheduler.step()

        kbar.update(i+1, values=list(loss_dict.items()))
    return


def auto_train(dataset_path, param_path, model_path):
    with open(param_path, "r") as f:
        params = yaml.load(f, Loader=yaml.SafeLoader)
    train(dataset_path, model_path, **params)


# assembling:
parser = argh.ArghParser()
parser.add_commands([
    auto_train,
])

# dispatching:
if __name__ == '__main__':
    parser.dispatch()

