import argh
import random
import shutil
import os
import json
from pycocotools.coco import COCO

random.seed(42)


@argh.arg('ratio', type=float)
def split_train_test(src_dir, dst_dir, ratio):
    """
    Splits images and masks to train and validation
    :param src_dir: directory containing subdirectories "images" and "annotations"
    :param dst_dir: directory path save the directories "train" and "test" into
    :param ratio: size of test set - test/total
    :return:
    """
    src_img_dir = os.path.join(src_dir, "images")
    src_mask_dir = os.path.join(src_dir, "annotations")
    dst_train_img_dir = os.path.join(dst_dir, "train", "images")
    dst_train_annotation_dir = os.path.join(dst_dir, "train", "annotations")
    dst_test_img_dir = os.path.join(dst_dir, "test", "images")
    dst_test_annotation_dir = os.path.join(dst_dir, "test", "annotations")
    for dst in [dst_train_img_dir, dst_train_annotation_dir, dst_test_img_dir, dst_test_annotation_dir]:
        os.makedirs(dst, exist_ok=True)

    train_dict = {"categories": [], "images": [], "annotations": []}
    test_dict = {"categories": [], "images": [], "annotations": []}
    for file_name in os.listdir(src_mask_dir):
        coco = COCO(os.path.join(src_mask_dir, file_name))
        train_dict["categories"] += coco.loadCats(coco.getCatIds())
        test_dict["categories"] += coco.loadCats(coco.getCatIds())
        for img_id in coco.getImgIds():
            images = coco.loadImgs([img_id])
            annotations = coco.loadAnns(coco.getAnnIds(imgIds=[img_id]))

            if random.uniform(0, 1) < ratio:
                test_dict["images"] += images
                test_dict["annotations"] += annotations
                img_dst = dst_test_img_dir
            else:
                train_dict["images"] += images
                train_dict["annotations"] += annotations
                img_dst = dst_train_img_dir

            img_file_name = images[0]["file_name"]
            shutil.copyfile(os.path.join(src_img_dir, img_file_name), os.path.join(img_dst, img_file_name))

    with open(os.path.join(dst_train_annotation_dir, "annotations.json"), "w") as f:
        json.dump(train_dict, f)

    with open(os.path.join(dst_test_annotation_dir, "annotations.json"), "w") as f:
        json.dump(test_dict, f)

    # load and display instance annotations
        # img_ids = coco.getImgIds()
        # img_dicts = coco.loadImgs(img_ids)
        #
        # for img_dict in img_dicts:
        #     img_path = os.path.join(data_dir, img_dict["path"])
        # img = Image.open(img_path)
        # plt.imshow(img); plt.axis('off')
        # annIds = coco.getAnnIds(imgIds=img_dict['id'], iscrowd=None)
        # anns = coco.loadAnns(annIds)
        # coco.showAnns(anns)
        # plt.savefig("preview/{}".format(os.path.basename(img_path)), bbox_inches='tight', pad_inches=0)
        # plt.clf()

        # bare_file_name = os.path.splitext(file_name)[0]
        # mask_file_name = next(file for file in os.listdir(src_mask_dir) if file.startswith(bare_file_name))
        # if mask_file_name is None:
        #     continue
        # if random.uniform(0, 1) < ratio:
        #     img_dst = dst_val_img_dir
        #     mask_dst = dst_val_mask_dir
        # else:
        #     img_dst = dst_train_img_dir
        #     mask_dst = dst_train_mask_dir
        #
        # shutil.copyfile(os.path.join(src_img_dir, file_name), os.path.join(img_dst, file_name))
        # shutil.copyfile(os.path.join(src_mask_dir, mask_file_name), os.path.join(mask_dst, mask_file_name))


# assembling:
parser = argh.ArghParser()
parser.add_commands([
    split_train_test,
])

# dispatching:
if __name__ == '__main__':
    parser.dispatch()
